"""Process the raw DEMoS dataset.

This assumes the file structure from the original compressed file:
/.../
    DEMOS/
        NP_*.wav
        PR_*.wav
    NEU/
        *.wav
    ...
"""

from pathlib import Path

import click

from ertk.dataset import resample_audio, write_annotations, write_filelist

emotion_map = {
    "rab": "anger",
    "tri": "sadness",
    "gio": "happiness",
    "pau": "fear",
    "dis": "disgust",
    "col": "guilt",
    "sor": "surprise",
    "neu": "neutral",
}


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the DEMoS dataset at location INPUT_DIR and resample
    audio to 16 kHz 16-bit WAV audio.
    """
    paths = list(input_dir.glob("**/*.wav"))
    resample_dir = Path("resampled")
    if resample:
        resample_audio(paths, resample_dir)
    write_filelist(resample_dir.glob("*.wav"), "files_all")
    write_filelist(resample_dir.glob("PR_*.wav"), "files_PR")
    write_filelist(resample_dir.glob("NP_*.wav"), "files_NP")

    write_annotations({p.stem: emotion_map[p.stem[-6:-3]] for p in paths}, "label")
    speaker_dict = {p.stem: p.stem[-9:-7] for p in paths}
    write_annotations(speaker_dict, "speaker")
    # fmt: off
    male_speakers = [
        "02", "03", "04", "05", "08", "09", "10", "11", "12", "14", "15", "16", "18",
        "19", "23", "24", "25", "26", "27", "28", "30", "33", "34", "39", "41", "50",
        "51", "52", "53", "58", "59", "63", "64", "65", "66", "67", "68", "69"
    ]
    # fmt: on
    gender_dict = {
        k: "M" if v in male_speakers else "F" for k, v in speaker_dict.items()
    }
    write_annotations(gender_dict, "gender")
    write_annotations({p.stem: "it" for p in paths}, "language")
    write_annotations({p.stem: "it" for p in paths}, "country")
    prot_dict = {
        p.stem: p.stem[:2] if p.stem[:2] in {"NP", "PR"} else "neutral" for p in paths
    }
    write_annotations(prot_dict, "prototypicality")


if __name__ == "__main__":
    main()



"""Implementation of auDeep model from [1]_.

References
----------
.. [1] M. Freitag, S. Amiriparian, S. Pugachevskiy, N. Cummins, and B.
       Schuller, “auDeep: Unsupervised learning of representations from
       audio with deep recurrent neural networks,” The Journal of
       Machine Learning Research, vol. 18, no. 1, pp. 6340-6344, 2017.
"""

import argparse
import time
from pathlib import Path

import netCDF4
import numpy as np
import tensorflow as tf
from keras import Model
from keras.layers import (
    RNN,
    AbstractRNNCell,
    Bidirectional,
    Dense,
    GRUCell,
    Input,
    StackedRNNCells,
    concatenate,
)
from keras.models import load_model
from tqdm import tqdm

__all__ = ["audeep_trae"]


def _dropout_gru_cell(units: int = 256, dropout: float = 0.8) -> AbstractRNNCell:
    return tf.nn.RNNCellDropoutWrapper(
        GRUCell(units, name="gru_cell"), 1 - dropout, 1 - dropout
    )


def _make_rnn(
    units: int = 256,
    layers: int = 2,
    bidirectional: bool = False,
    dropout: float = 0.2,
    name="rnn",
) -> RNN:
    cells = [_dropout_gru_cell(units, dropout) for _ in range(layers)]
    stacked_cell = StackedRNNCells(cells)
    rnn = RNN(
        stacked_cell,
        return_sequences=True,
        return_state=True,
        time_major=True,
        name=name,
    )
    if bidirectional:
        rnn = Bidirectional(rnn, name="bidirectional_" + name)
    return rnn


def audeep_trae(
    input_shape: tuple,
    units: int = 256,
    layers: int = 2,
    bidirectional_encoder: bool = False,
    bidirectional_decoder: bool = False,
) -> Model:
    inputs = Input(input_shape)

    trans = tf.transpose(inputs, [1, 0, 2])

    # Make encoder layers
    encoder = _make_rnn(units, layers, bidirectional_encoder, name="encoder")
    enc_states = encoder(trans)[1:]
    encoder_output = concatenate(enc_states, name="encoder_output_state")

    # Fully connected needs to have output dimension equal to dimension of
    # decoder state
    dec_dim = units * layers
    if bidirectional_decoder:
        dec_dim *= 2
    representation = Dense(dec_dim, activation="tanh", name="representation")(
        encoder_output
    )

    # Initial state of decoder
    decoder_init_state = tf.split(
        representation,
        2 * layers if bidirectional_decoder else layers,
        axis=-1,
        name="decoder_init_state",
    )
    # Decoder input is reversed and shifted input sequence
    targets = trans[:0:-1]
    dec_inputs = tf.pad(
        targets, [[1, 0], [0, 0], [0, 0]], name="decoder_input_sequence"
    )

    # Make decoder layers and init with output from fully connected layer
    decoder = _make_rnn(units, layers, bidirectional_decoder, name="decoder")
    decoder_output = decoder(dec_inputs, initial_state=decoder_init_state)[0]

    n_features = input_shape[-1]
    reconstruction = Dense(n_features, activation="tanh", name="reconstruction")(
        decoder_output
    )

    reconstruction = tf.transpose(reconstruction, [1, 0, 2])

    model = Model(inputs=inputs, outputs=[reconstruction, representation])
    return model


def _make_functions(model, optimizer, strategy, use_function=True):
    def train_step(data):
        with tf.GradientTape() as tape:
            reconstruction, _ = model(data, training=True)
            targets = data[:, ::-1, :]
            loss = tf.sqrt(
                tf.reduce_mean(tf.math.squared_difference(targets, reconstruction))
            )

        trainable_vars = model.trainable_variables
        gradients = tape.gradient(loss, trainable_vars)
        clipped_gvs = [tf.clip_by_value(x, -2, 2) for x in gradients]
        # clipped_gvs, _ = tf.clip_by_global_norm(gradients, 2)
        optimizer.apply_gradients(zip(clipped_gvs, trainable_vars))

        return loss

    def test_step(data):
        reconstruction, _ = model(data, training=False)
        targets = data[:, ::-1, :]
        loss = tf.sqrt(
            tf.reduce_mean(tf.math.squared_difference(targets, reconstruction))
        )

        return loss

    def dist_train_step(batch):
        per_replica_losses = strategy.run(train_step, args=(batch,))
        return strategy.reduce(
            tf.distribute.ReduceOp.SUM, per_replica_losses, axis=None
        )

    def dist_test_step(batch):
        per_replica_losses = strategy.run(test_step, args=(batch,))
        return strategy.reduce(
            tf.distribute.ReduceOp.MEAN, per_replica_losses, axis=None
        )

    if use_function:
        dist_train_step = tf.function(dist_train_step)
        dist_test_step = tf.function(dist_test_step)

    return dist_train_step, dist_test_step


# TODO: Update to use ertk.dataset.Dataset, etc.
def train(args):
    args.logs.parent.mkdir(parents=True, exist_ok=True)
    train_log_dir = str(args.logs / "train")
    valid_log_dir = str(args.logs / "valid")
    train_writer = tf.summary.create_file_writer(train_log_dir)
    valid_writer = tf.summary.create_file_writer(valid_log_dir)

    if args.profile:
        tf.profiler.experimental.start("")
        tf.profiler.experimental.stop(save=False)

    strategy = tf.distribute.get_strategy()
    if args.multi_gpu:
        strategy = tf.distribute.MirroredStrategy()

    # Get data
    dataset = netCDF4.Dataset(str(args.dataset))
    x = np.array(dataset.variables["features"])
    dataset.close()
    n_spectrograms, max_time, features = x.shape
    np.random.default_rng().shuffle(x)
    n_valid = int(n_spectrograms * args.valid_fraction)
    valid_data = (
        tf.data.Dataset.from_tensor_slices(x[:n_valid])
        .batch(args.batch_size)
        .prefetch(10)
    )
    train_data = (
        tf.data.Dataset.from_tensor_slices(x[n_valid:])
        .shuffle(n_spectrograms)
        .batch(args.batch_size)
        .prefetch(100)
    )
    summary_data = valid_data.take(1)
    # Reduce memory footprint
    del x

    train_data = strategy.experimental_distribute_dataset(train_data)
    valid_data = strategy.experimental_distribute_dataset(valid_data)

    n_train_batches = len(train_data)
    n_valid_batches = len(valid_data)

    step = tf.Variable(1)
    with strategy.scope():
        model = audeep_trae(
            (max_time, features),
            units=args.units,
            layers=args.layers,
            bidirectional_encoder=args.bidirectional_encoder,
            bidirectional_decoder=args.bidirectional_decoder,
        )
        # Using epsilon=1e-5 seems to offer better stability.
        optimizer = tf.optimizers.Adam(learning_rate=args.learning_rate, epsilon=1e-5)
        model.compile(optimizer=optimizer)
        checkpoint = tf.train.Checkpoint(model=model, optimizer=optimizer, step=step)
    print()
    print("Model structure:")
    for layer in model.layers:
        name = layer.name
        if name.startswith("tf_op_layer_"):
            name = name[12:]
        print(layer.name, layer.output_shape)
    print()

    # Write the graph only
    tf.keras.callbacks.TensorBoard(
        args.logs / "graph", write_graph=True, profile_batch=0
    ).set_model(model)

    checkpoint_manager = tf.train.CheckpointManager(
        checkpoint,
        str(args.logs),
        max_to_keep=args.keep_checkpoints,
        checkpoint_interval=args.save_interval,
        step_counter=step,
    )
    if args.cont:
        checkpoint.restore(checkpoint_manager.latest_checkpoint)
        print(f"Restoring from checkpoint " f"{checkpoint_manager.latest_checkpoint}")

    train_step, test_step = _make_functions(model, optimizer, strategy)

    start_epoch = step.value().numpy()
    for epoch in range(start_epoch, start_epoch + args.epochs):
        step.assign(epoch)
        epoch_time = time.perf_counter()
        train_loss = 0.0
        n_batch = 0
        batch_time = time.perf_counter()
        for batch in tqdm(
            train_data, total=n_train_batches, desc=f"Epoch {epoch:03d}", unit="batch"
        ):
            if epoch == 2 and n_batch == 0 and args.profile:
                tf.profiler.experimental.start(args.logdir)
            loss = train_step(batch)
            train_loss += loss
            n_batch += 1
            if epoch == 2 and n_batch == 0 and args.profile:
                tf.profiler.experimental.stop()
        train_loss /= n_train_batches
        batch_time = time.perf_counter() - batch_time
        batch_time /= n_train_batches

        valid_loss = 0.0
        for batch in valid_data:
            loss = test_step(batch)
            valid_loss += loss
        valid_loss /= n_valid_batches

        epoch_time = time.perf_counter() - epoch_time
        print(
            f"Epoch {epoch:03d}: train loss: {train_loss:.4f}, valid loss: "
            f"{valid_loss:.4f} in {epoch_time:.2f}s ({batch_time:.2f}s per "
            "batch)."
        )

        with valid_writer.as_default():
            tf.summary.scalar("rmse", valid_loss, step=epoch)

            batch = next(iter(summary_data))
            reconstruction, representation = model(batch, training=False)
            reconstruction = reconstruction[:, ::-1, :]
            images = tf.concat([batch, reconstruction], 2)
            images = tf.expand_dims(images, -1)
            images = (images + 1) / 2
            tf.summary.image("combined", images, step=epoch, max_outputs=20)
            tf.summary.histogram("representation", representation, step=epoch)

        with train_writer.as_default():
            tf.summary.scalar("rmse", train_loss, step=epoch)

        if epoch % args.save_interval == 0:
            checkpoint_manager.save(checkpoint_number=epoch)
            print(f"Saved checkpoint to " f"{checkpoint_manager.latest_checkpoint}")

    save_path = str(args.logs / "model")
    model.save(save_path, include_optimizer=False)
    print(f"Saved model to {save_path}")


def generate(args):
    dataset = netCDF4.Dataset(args.dataset)
    data = tf.data.Dataset.from_tensor_slices(dataset.variables["features"]).batch(
        args.batch_size
    )
    filenames = np.array(dataset.variables["name"])
    labels = np.array(dataset.variables["label"])
    corpus = dataset.corpus
    dataset.close()

    model = load_model(args.model)
    # Optimise model call function
    model.call = tf.function(model.call)

    print(f"Read dataset from {args.dataset}")

    representations = []
    for batch in tqdm(data):
        _, representation = model(batch, training=False)
        representations.append(representation.numpy())
    representations = np.concatenate(representations)

    dataset = netCDF4.Dataset(str(args.output), "w")
    dataset.createDimension("instance", len(filenames))
    dataset.createDimension("generated", representations.shape[-1])

    filename = dataset.createVariable("name", str, ("instance",))
    filename[:] = filenames

    label_nominal = dataset.createVariable("label", str, ("instance",))
    label_nominal[:] = labels

    features = dataset.createVariable("features", np.float32, ("instance", "generated"))
    features[:, :] = representations

    dataset.setncattr_string("feature_dims", '["generated"]')
    dataset.setncattr_string("corpus", corpus)
    dataset.close()

    print(f"Wrote netCDF4 file to {args.output}")


def main():
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(title="command", dest="command", required=True)

    train_args = subparsers.add_parser("train")
    train_args.add_argument(
        "--dataset", type=Path, required=True, help="File containing spectrogram data."
    )
    train_args.add_argument(
        "--logs", type=Path, required=True, help="Directory to store TensorBoard logs."
    )

    train_args.add_argument(
        "--epochs", type=int, default=50, help="Number of epochs to train for."
    )
    train_args.add_argument(
        "--valid_fraction",
        type=float,
        default=0.1,
        help="Fraction of data to use as validation data.",
    )
    train_args.add_argument(
        "--learning_rate", type=float, default=0.001, help="Learning rate."
    )
    train_args.add_argument(
        "--batch_size", type=int, default=64, help="Global batch size."
    )
    train_args.add_argument(
        "--multi_gpu", action="store_true", help="Use all GPUs for training."
    )
    train_args.add_argument("--profile", action="store_true", help="Profile a batch.")
    train_args.add_argument(
        "--continue",
        action="store_true",
        dest="cont",
        help="Continue from latest checkpoint.",
    )
    train_args.add_argument(
        "--keep_checkpoints", type=int, default=5, help="Number of checkpoints to keep."
    )
    train_args.add_argument(
        "--save_interval", type=int, default=10, help="Save every N epochs."
    )

    train_args.add_argument(
        "--units", type=int, default=256, help="Dimensionality of RNN cells."
    )
    train_args.add_argument(
        "--layers", type=int, default=2, help="Number of stacked RNN layers."
    )
    train_args.add_argument(
        "--bidirectional_encoder",
        action="store_true",
        help="Use a bidirectional encoder.",
    )
    train_args.add_argument(
        "--bidirectional_decoder",
        action="store_true",
        help="Use a bidirectional decoder.",
    )

    gen_args = subparsers.add_parser("generate")
    gen_args.add_argument(
        "--dataset", type=Path, required=True, help="File containing spectrogram data."
    )
    gen_args.add_argument(
        "--model", type=Path, required=True, help="Directory of Keras saved_model."
    )
    gen_args.add_argument("--batch_size", type=int, default=256, help="Batch size.")
    gen_args.add_argument(
        "--output", type=Path, required=True, help="Path to output representations."
    )

    args = parser.parse_args()

    for gpu in tf.config.list_physical_devices("GPU"):
        tf.config.experimental.set_memory_growth(gpu, True)

    if args.command == "train":
        train(args)
    elif args.command == "generate":
        generate(args)


if __name__ == "__main__":
    main()



"""OpenXBOW feature extractor."""

import os
import subprocess
import tempfile
from dataclasses import dataclass
from importlib.resources import path as res_path
from typing import Iterable, List, Union

import numpy as np
import pandas as pd
from omegaconf import MISSING

from ertk.config import ERTKConfig
from ertk.dataset import write_features
from ertk.preprocessing._base import FeatureExtractor

__all__ = ["OpenXBOWExtractorConfig", "OpenXBOWExtractor"]


@dataclass
class OpenXBOWExtractorConfig(ERTKConfig):
    """Configuration for the OpenXBOW feature extractor."""

    xbowargs: List[str] = MISSING
    """The arguments to pass to OpenXBOW."""


class OpenXBOWExtractor(
    FeatureExtractor, fname="openxbow", config=OpenXBOWExtractorConfig
):
    """OpenXBOW feature extractor."""

    config: OpenXBOWExtractorConfig
    _dim = None

    def __init__(self, config: OpenXBOWExtractorConfig) -> None:
        super().__init__(config)

    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        raise NotImplementedError("Use batch_size=-1 for openxbow.")

    def process_batch(
        self, batch: Union[Iterable[np.ndarray], np.ndarray], **kwargs
    ) -> List[np.ndarray]:
        _, tmpin = tempfile.mkstemp(prefix="openxbow_", suffix=".csv")
        _, tmpout = tempfile.mkstemp(prefix="openxbow_", suffix=".csv")

        xs = list(batch)
        names = [str(i) for i in range(len(xs))]
        self.logger.info(f"Writing temp CSV to {tmpin}")
        write_features(tmpin, xs, names=names, header=False)

        add_args = []
        for arg in self.config.xbowargs:
            add_args.extend(arg.split("=", maxsplit=1))

        with res_path("ertk.preprocessing.openxbow", "openXBOW.jar") as jar:
            xbow_args = [
                "java",
                "-jar",
                str(jar),
                "-i",
                tmpin,
                "-o",
                tmpout,
                "-attributes",
                f"n1[{xs[0].shape[1]}]",
                "-csvSep",
                ",",
                "-writeName",
                "-noLabels",
                *add_args,
            ]

        self.logger.info(f"Using cmdline: {' '.join(xbow_args)}")
        if subprocess.call(xbow_args) != 0:
            raise RuntimeError("openXBOW returned an error")
        os.remove(tmpin)
        data: pd.DataFrame = pd.read_csv(
            tmpout, header=None, quotechar="'", dtype={0: str}
        )
        os.remove(tmpout)
        self._dim = data.shape[1] - 1
        self.logger.debug(data.shape)
        if not len(data) == len(xs):
            raise RuntimeError("Got incorrect number of instances from openXBOW.")

        return list(data.iloc[:, 1:].to_numpy())

    @property
    def dim(self) -> int:
        for i, arg in enumerate(self.config.xbowargs):
            if arg.startswith("-size"):
                try:
                    _, size = arg.split("=", maxsplit=1)
                except ValueError:
                    size = self.config.xbowargs[i + 1]
                return int(size)
        if self._dim:
            return self._dim
        raise ValueError("Cannot determine dim.")

    @property
    def is_sequence(self) -> bool:
        return False

    @property
    def feature_names(self) -> List[str]:
        return [f"bow_{i}" for i in range(self.dim)]



"""
Configuration utilities
=======================

This module provides utilities for working with configuration
dataclasses and YAML files.


Base classes
------------
.. autosummary::
    :toctree: generated/

    ERTKConfig


Functions
---------
.. autosummary::
    :toctree: generated/

    get_arg_mapping
"""

from abc import ABC
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Type, TypeVar, Union, cast

import yaml
from omegaconf import DictConfig, OmegaConf

from ertk.utils import PathOrStr

__all__ = ["ERTKConfig", "get_arg_mapping"]


T = TypeVar("T", bound="ERTKConfig")


def resolve_files(key):
    return key


def resolve_files_load(key):
    return OmegaConf.load(key)


OmegaConf.register_resolver("file", resolve_files)
OmegaConf.register_resolver("cwdpath", resolve_files_load)


@dataclass
class ERTKConfig(ABC):
    """Base class for ERTK configuration dataclasses."""

    def to_dictconfig(self) -> DictConfig:
        """Convert config to DictConfig.

        Returns
        -------
        omegaconf.DictConfig
            The structured config which behaves like the corresponding
            dataclass.
        """
        return DictConfig(self)

    @classmethod
    def from_config(cls: Type[T], config: Any) -> T:
        """Create config object from any compatible config.

        Parameters
        ----------
        config: omegaconf.DictConfig or dict
            The object to create the config from.

        Returns
        -------
        ERKConfig
            The resulting config.
        """
        schema = OmegaConf.structured(cls)
        return cast(T, OmegaConf.merge(schema, config))

    @classmethod
    def from_file(
        cls: Type[T], path: PathOrStr, override: Optional[List[str]] = None
    ) -> T:
        """Create config from YAML file and optionlly override some
        values.

        Parameters
        ----------
        path: os.Pathlike or str
            The path to YAML file containing config.
        override: list of str, optional
            Argument overrides in the form of key=value pairs.

        Returns
        -------
        ERKConfig
            The resulting config.
        """
        config = OmegaConf.load(Path(path))
        if override is not None:
            config = OmegaConf.merge(config, OmegaConf.from_dotlist(override))
        return cls.from_config(config)

    def to_string(self) -> str:
        """Generate YAML string representation of config.

        Returns
        -------
        str
            YAML string representation of config.
        """
        return OmegaConf.to_yaml(self)

    def to_file(self, path: PathOrStr) -> None:
        """Write config to YAML file.

        Parameters
        ----------
        path: os.Pathlike or str
            The path to YAML file to write config to.
        """
        with open(path, "w") as fid:
            fid.write(OmegaConf.to_yaml(self))

    def merge_with_args(self: T, args: Optional[List[str]] = None) -> T:
        """Merge config with command-line arguments.

        Parameters
        ----------
        args: list of str, optional
            Argument overrides in the form of key=value pairs.

        Returns
        -------
        ERKConfig
            The resulting config.
        """
        return cast(T, OmegaConf.merge(self, OmegaConf.from_cli(args)))


def get_arg_mapping(s: Union[Path, str]) -> Dict[str, Any]:
    """Given a mapping on the command-line, returns a dict representing
    that mapping. Mapping can be a string or a more complex YAML file.

    The string form of the mapping is:
        key:value[,key:value]+

    Parameters
    ----------
    s: PathLike or str
        String representing the mapping or path to YAML containing
        mapping. If a string, it cannot contain spaces or shell symbols
        (unless escaped).

    Returns
    -------
    mapping: dict
        A dictionary mapping keys to values from the string.
    """
    if isinstance(s, Path) or Path(s).exists():
        with open(s) as fid:
            return yaml.safe_load(fid) or {}
    mapping: Dict[str, List[str]] = defaultdict(list)
    for cls in s.split(","):
        key, val = cls.split(":")
        mapping[key].append(val)
    return {k: v[0] if len(v) == 1 else v for k, v in mapping.items()}



"""
Preprocessor classes and plugins
================================

This module contains base classes for audio clip processors, feature
extractors and instance processors, as well as a number of plugins for
each of these.


Base classes
------------
.. autosummary::
    :toctree: generated/

    InstanceProcessor
    AudioClipProcessor
    FeatureExtractor


Plugins
-------
.. autosummary::
    :toctree: generated/

    audioset
    encodec
    fairseq
    huggingface
    keras_apps
    kmeans
    opensmile
    openxbow
    phonemize
    resample
    spectrogram
    speechbrain
    vad_trim
"""

from ._base import AudioClipProcessor, FeatureExtractor, InstanceProcessor



"""Process the raw OréauFR dataset.

This assumes the file structure from the original compressed file:
/.../
    OréauFR_01/
        m/
            sessc/
                *.wav
            ...
        f/
        contents.txt
    OréauFR_02/
    ...
"""

from pathlib import Path

import click

from ertk.dataset import resample_audio, write_annotations, write_filelist

emotion_map = {
    "C": "anger",
    "D": "disgust",
    "J": "happiness",
    "N": "neutral",
    "P": "fear",
    "S": "surprise",
    "T": "sadness",
}


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process OréauFR dataset at location INPUT_DIR."""

    paths = list(input_dir.glob("OréauFR_01/*/*/*.wav"))
    if resample:
        resample_dir = Path("resampled")
        resample_dir.mkdir(exist_ok=True)
        resample_audio(paths, resample_dir)
        paths = list(resample_dir.glob("*.wav"))
    write_filelist(paths, "files_all")
    write_filelist(paths, "files_01")
    write_filelist(input_dir.glob("OréauFR_02/*/*/*.wav"), "files_02")

    write_annotations({p.stem: emotion_map[p.stem[5]] for p in paths}, "label")
    write_annotations({p.stem: p.stem[0:2] for p in paths}, "speaker")
    write_annotations({p.stem: p.stem[2:5] for p in paths}, "sentence")
    write_annotations({p.stem: "fr" for p in paths}, "language")


if __name__ == "__main__":
    main()



"""Process the raw eNTERFACE dataset.

This assumes the file structure from the original compressed file:
/.../
    subject 1/
        anger/
            sentence 1/
                *.avi
            ...
        ...
    ...
"""

import shutil
from pathlib import Path

import click

from ertk.dataset import resample_audio, write_annotations, write_filelist

emotion_map = {
    "an": "anger",
    "di": "disgust",
    "fe": "fear",
    "ha": "happiness",
    "sa": "sadness",
    "su": "surprise",
}


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the eNTERFACE dataset at location INPUT_DIR and convert
    AVI to 16 kHz 16-bit WAV audio.
    """
    paths = list(input_dir.glob("**/*.avi"))
    resample_dir = Path("resampled")

    if resample:
        # Correct duplicate names for subject 11 separately
        sub11 = [p for p in paths if "subject 11" in p.parts]
        resample_audio(sub11, resample_dir)
        for f in resample_dir.glob("s12*"):
            shutil.move(f, f.with_name(f.name.replace("s12", "s11")))

        rest = [p for p in paths if "subject 11" not in p.parts]
        resample_audio(rest, resample_dir)

        # More manual corrections
        shutil.move("resampled/s16_su_3avi.wav", "resampled/s16_su_3.wav")
        for f in resample_dir.glob("s_3_*"):
            shutil.move(f, f.with_name(f.name.replace("s_3", "s3")))

    newpaths = list(resample_dir.glob("*.wav"))
    write_filelist([p for p in newpaths], "files_all")
    write_filelist([p for p in newpaths if not p.stem.startswith("s6_")], "files_no_s6")
    write_annotations(
        {
            p.stem: emotion_map[p.stem[p.stem.find("_") + 1 : p.stem.find("_") + 3]]
            for p in newpaths
        },
        "label",
    )
    write_annotations({p.stem: p.stem[: p.stem.find("_")] for p in newpaths}, "speaker")
    write_annotations({p.stem: "en" for p in newpaths}, "language")


if __name__ == "__main__":
    main()



"""Scikit-learn classification functions."""

import json
import logging
import time
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
from sklearn.base import ClassifierMixin
from sklearn.metrics import confusion_matrix
from sklearn.model_selection import BaseCrossValidator, cross_validate
from sklearn.model_selection._search import BaseSearchCV
from sklearn.pipeline import Pipeline

from ertk.sklearn.utils import get_base_estimator, get_estimator_tree
from ertk.train import ExperimentResult, get_pipeline_params, get_scores
from ertk.utils import ScoreFunction

__all__ = ["sk_cross_validate", "sk_train_val_test"]

logger = logging.getLogger("ertk.sklearn.classification")


def sk_cross_validate(
    clf: ClassifierMixin,
    x: np.ndarray,
    y: Optional[np.ndarray] = None,
    *,
    groups: Optional[np.ndarray] = None,
    cv: BaseCrossValidator = None,
    scoring: Union[
        str, List[str], Dict[str, ScoreFunction], Callable[..., float]
    ] = "accuracy",
    verbose: int = 0,
    n_jobs: int = 1,
    fit_params: Dict[str, Any] = {},
) -> ExperimentResult:
    logger.debug(f"cross_validate(): filtered fit_params={fit_params}")
    logger.info(f"cross_validate(): verbose={verbose}")

    scores = cross_validate(
        clf,
        x,
        y,
        cv=cv,
        scoring=scoring,
        groups=groups,
        n_jobs=n_jobs,
        verbose=verbose,
        fit_params=fit_params,
        error_score="raise",
        return_estimator=True,
    )
    params = []
    for est in scores["estimator"]:
        if isinstance(est, BaseSearchCV):
            params.append(
                json.dumps(get_base_estimator(est.best_estimator_).get_params())
            )
        else:
            params.append(json.dumps(get_base_estimator(est).get_params()))
    scores["params"] = np.array(params)
    del scores["estimator"]
    return ExperimentResult(scores)


def sk_train_val_test(
    clf: ClassifierMixin,
    train_data: Tuple[np.ndarray, ...],
    valid_data: Tuple[np.ndarray, ...],
    test_data: Optional[Tuple[np.ndarray, ...]] = None,
    verbose: int = 0,
    scoring: Union[
        str, List[str], Dict[str, ScoreFunction], Callable[..., float]
    ] = "accuracy",
    fit_params: Dict[str, Any] = {},
) -> ExperimentResult:
    sw_kw = {"sample_weight": train_data[2] if len(train_data) > 2 else None}
    for est in get_estimator_tree(clf):
        if isinstance(est, Pipeline):
            sw_kw = get_pipeline_params(sw_kw, est)
            break
    # TODO: make array arg lengths the same
    fit_params = dict(**fit_params, **sw_kw)
    logger.debug(f"train_val_test(): filtered fit_params={fit_params}")

    logger.info(f"Fitting classifier {clf}")
    start_time = time.perf_counter()
    clf = clf.fit(train_data[0], train_data[1], **fit_params)
    fit_time = time.perf_counter() - start_time
    logger.info(f"Getting predictions from classifier {clf}")
    start_time = time.perf_counter()
    y_pred = clf.predict(test_data[0])
    score_time = time.perf_counter() - start_time
    scores: Dict[str, Any] = {
        "fit_time": fit_time,
        "score_time": score_time,
        **get_scores(scoring, y_pred, test_data[1]),
    }
    scores = {f"test_{k}": v for k, v in scores.items()}
    print(confusion_matrix(test_data[1], y_pred))
    if isinstance(clf, BaseSearchCV):
        scores["params"] = json.dumps(
            get_base_estimator(clf.best_estimator_).get_params()
        )
    else:
        scores["params"] = json.dumps(get_base_estimator(clf).get_params())
    return ExperimentResult(scores, predictions=y_pred)



"""Fairseq processor."""

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional, Union

import joblib
import numpy as np
import torch
import torch.nn.functional as F
from omegaconf import MISSING, OmegaConf

from ertk.config import ERTKConfig
from ertk.utils import PathOrStr, is_mono_audio

from ._base import AudioClipProcessor, FeatureExtractor

__all__ = ["FairseqExtractorConfig", "FairseqExtractor"]


class Agg(Enum):
    """Aggregation method."""

    MEAN = "mean"
    """Mean of the output sequence."""
    MAX = "max"
    """Max of the output sequence."""
    NONE = "none"
    """Return entire sequence."""


@dataclass
class FairseqExtractorConfig(ERTKConfig):
    """Fairseq feature extractor configuration."""

    model_type: str = MISSING
    """Model type."""
    checkpoint: str = MISSING
    """Path to model checkpoint."""
    layer: str = "context"
    """Layer to extract features from."""
    aggregate: Agg = Agg.MEAN
    """Aggregation method."""
    device: str = "cuda"
    """Device to run model on."""
    arg_overrides: Dict[str, Any] = field(default_factory=dict)
    """Overrides for model arguments."""
    vq_path: Optional[str] = None
    """Path to vector quantiser."""
    vq_ids: bool = False
    """Whether to return VQ cluster ids."""
    vq_ids_as_string: bool = True
    """Whether to return VQ cluster ids as a single string of integers
    separated by spaces.
    """
    max_input_len: int = 1500000
    """Maximum input length."""
    mms_lang: str = "eng"
    """Language to use for MMS models."""


class FairseqExtractor(
    FeatureExtractor, AudioClipProcessor, fname="fairseq", config=FairseqExtractorConfig
):
    """Fairseq processor."""

    config: FairseqExtractorConfig

    def __init__(self, config: FairseqExtractorConfig) -> None:
        super().__init__(config)

        print(f"Loading model from {config.checkpoint}")
        if self.config.model_type == "data2vec":
            # Have to do everything manually since the released model is
            # 'experimental'
            import fairseq.examples.data2vec.models.data2vec_audio  # noqa
            from fairseq import models, tasks

            state = torch.load(config.checkpoint)
            cfg = OmegaConf.create(state["cfg"])

            del cfg.task["precompute_mask_indices"]
            del cfg.task["inferred_w2v_config"]
            task = tasks.setup_task(cfg.task)

            _model_dataclass = models.MODEL_DATACLASS_REGISTRY["data2vec_audio"]
            for key in set(cfg.model):
                if key not in _model_dataclass.__dataclass_fields__:
                    del cfg.model[key]
            model = task.build_model(cfg.model)

            del state["model"]["_ema"]
            state["model"]["final_proj.weight"] = state["model"].pop(
                "final_proj.0.weight"
            )
            state["model"]["final_proj.bias"] = state["model"].pop("final_proj.0.bias")
            model.load_state_dict(state["model"])

            args = OmegaConf.create(state["args"])
        elif self.config.model_type == "mms":
            from fairseq.checkpoint_utils import load_checkpoint_to_cpu
            from fairseq.tasks import setup_task

            ckpt = load_checkpoint_to_cpu(config.checkpoint)
            assert config.mms_lang in ckpt["adapter"]

            cfg = ckpt["cfg"]
            args = cfg
            task = setup_task(cfg.task, from_checkpoint=True)
            task.load_state_dict(ckpt["task_state"])  # Needed to build model
            model = task.build_model(cfg.model, from_checkpoint=True)
            model.load_state_dict(ckpt["model"], strict=True, model_cfg=cfg.model)

            # Load the correct language adapter
            ft_obj = ckpt["adapter"][config.mms_lang]
            ft_model = ft_obj["model"]
            cdevice = model.w2v_encoder.proj.weight.device
            cdtype = model.w2v_encoder.proj.weight.dtype
            proj_out, proj_in = ft_model["w2v_encoder.proj.weight"].shape
            ft_proj = torch.nn.Linear(proj_in, proj_out, bias=True)
            ft_proj.to(device=cdevice, dtype=cdtype)
            model.w2v_encoder.proj = ft_proj
            with torch.no_grad():
                for kk, vv in model.named_parameters():
                    if kk in ft_model:
                        vv.copy_(ft_model[kk])
            task.load_state_dict(ft_obj["task_state"])
        else:
            from fairseq.checkpoint_utils import load_model_ensemble_and_task

            [model], args, task = load_model_ensemble_and_task(
                [config.checkpoint], arg_overrides=config.arg_overrides
            )
        model.to(device=config.device)
        model.eval()
        self.model = model
        self.args = args
        self.task = task
        torch.set_grad_enabled(False)

        self.vq = None
        if config.vq_path:
            self.vq = joblib.load(config.vq_path)

    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        if kwargs.pop("sr") != 16000:
            raise ValueError("Sample rate should be 16kHz")

        x = np.squeeze(x)
        if not is_mono_audio(x):
            raise ValueError("Audio must be mono")
        x = x[: self.config.max_input_len]
        tensor = torch.tensor(
            x, dtype=torch.float32, device=self.config.device
        ).unsqueeze(0)
        if hasattr(self.task, "normalize") and self.task.normalize:
            tensor = F.layer_norm(tensor, tensor.shape)

        layer: Union[str, int]
        try:
            layer = int(self.config.layer)
        except ValueError:
            layer = self.config.layer
        if layer == "context":
            layer = -1

        # Automatic speech recognition
        if self.config.model_type == "mms":
            outputs = self.model.forward(
                source=tensor, padding_mask=None, corpus_key=[0]
            )
            logprobs = self.model.get_normalized_probs(outputs, log_probs=True)
            toks = logprobs.transpose(0, 1).argmax(dim=-1).cpu().unique_consecutive()
            toks = toks[toks != 0]
            sent: str = self.task.target_dictionary.string(toks.long())
            sent = sent.replace(" ", "").replace("|", " ").strip()
            return np.array([sent], dtype=object)

        # Embeddings
        if self.config.model_type == "wav2vec":
            feats = self.model.feature_extractor(tensor)
            if self.model.vector_quantizer is not None:
                feats, _ = self.model.vector_quantizer.forward_idx(feats)
            if layer == -1:
                feats = self.model.feature_aggregator(feats)
        elif self.config.model_type in ["wav2vec2", "data2vec"]:
            if layer == "encoder":
                feats = self.model.feature_extractor(tensor)
            else:
                c = self.model.extract_features(tensor, None, layer=layer)["x"]
                # Transpose to (batch, feats, steps)
                feats = c.transpose(-2, -1)
        elif self.config.model_type == "hubert":
            if layer == "encoder":
                feats = self.model.feature_extractor(tensor)
            else:
                c, _ = self.model.extract_features(tensor, output_layer=layer)
                # Transpose to (batch, feats, steps)
                feats = c.transpose(-2, -1)
        else:
            raise ValueError(f"Unknown model_type: {self.config.model_type}")

        if self.config.aggregate == Agg.NONE:
            feats = feats[0].transpose(1, 0).cpu().numpy()
        elif self.config.aggregate == Agg.MEAN:
            feats = feats[0].mean(-1).cpu().numpy()
        elif self.config.aggregate == Agg.MAX:
            feats = feats[0].max(-1).cpu().numpy()
        else:
            raise ValueError(f"Aggregation type {self.config.aggregate} not supported")

        if self.vq is None:
            return feats

        # Vector quantization
        if self.config.vq_ids:
            feats = self.vq.predict(feats)
            if self.config.vq_ids_as_string:
                feats = np.array([" ".join(map(str, feats))])
            feats = np.expand_dims(feats, -1)
        else:
            feats = self.vq.cluster_centers_[self.vq.predict(feats)]
        return feats

    def process_file(self, path: PathOrStr, sr: Optional[float] = None) -> np.ndarray:
        # Require 16 kHz sample rate
        return super().process_file(path, 16000)

    @property
    def dim(self) -> int:
        if self.config.vq_ids or self.config.model_type == "mms":
            return 1
        if self.config.layer == "encoder":
            return 512
        if self.config.model_type == "wav2vec":
            return 512
        elif self.config.model_type in ["wav2vec2", "data2vec", "hubert"]:
            return self.model.encoder.embedding_dim
        raise ValueError("Unknown dim")

    @property
    def is_sequence(self) -> bool:
        return self.config.aggregate == Agg.NONE

    @property
    def feature_names(self) -> List[str]:
        if self.config.vq_ids:
            return [f"{self.config.model_type}_vq_tok"]
        if self.config.model_type == "mms":
            return ["text"]
        return [f"{self.config.model_type}_{i}" for i in range(self.dim)]



from dataclasses import dataclass

import omegaconf
import torch
import torch.nn.functional as F
from fairseq.checkpoint_utils import load_model_ensemble_and_task
from torch import nn

from ._base import PyTorchModelConfig, SimpleClassificationModel


@dataclass
class Wav2VecFineTuneConfig(PyTorchModelConfig):
    checkpoint: str = omegaconf.MISSING
    frozen_layers: int = -1


class Model(
    SimpleClassificationModel, fname="wav2vec_ft", config=Wav2VecFineTuneConfig
):
    def __init__(
        self,
        config: Wav2VecFineTuneConfig,
    ) -> None:
        super().__init__(config)

        [model], args, task = load_model_ensemble_and_task([config.checkpoint])
        self.wav2vec = model
        self.w2v_args = args
        # self.w2v_task = task

        # Needed for multi-GPU training because of
        # https://github.com/pytorch/fairseq/issues/3482
        self.w2v_cfg = dict(task.cfg)
        if "text_compression_level" in self.w2v_cfg:
            self.w2v_cfg["text_compression_level"] = str(
                self.w2v_cfg["text_compression_level"]
            )

        self.final = nn.Linear(512, self.n_classes)

    def forward(self, x):
        x = x.squeeze(-1)
        x = self.wav2vec.feature_extractor(x)
        if self.wav2vec.vector_quantizer is not None:
            x, _ = self.wav2vec.vector_quantizer.forward_idx(x)
        x = x.mean(-1)
        x = self.final(x)
        return x

    def preprocess_input(self, x: torch.Tensor) -> torch.Tensor:
        if self.w2v_cfg["normalize"]:
            with torch.no_grad():
                x = F.layer_norm(x, x.shape)
        return x



"""Process the raw MSP-IMPROV dataset.

This assumes the file structure from the original compressed file:
/.../
    Evaluation.txt
    session1/
        S01A/
            P/
                *.avi
                *.wav
            ...
        ...
    ...
"""


import re
from pathlib import Path

import click
import numpy as np
import pandas as pd

from ertk.dataset import resample_audio, write_annotations, write_filelist
from ertk.stats import alpha

REGEX = re.compile(
    r"^UTD-IMPROV-([A-Z0-9-]+)\.avi; ([A-Z]); A:(\d+\.\d+|NaN); V:(\d+\.\d+|NaN); D:(\d+\.\d+|NaN) ; N:(\d+\.\d+|NaN);$"  # noqa
)
RATER_RE = re.compile(
    r"^([A-Za-z0-9-_]+); (Sad|Angry|Happy|Neutral|Other); (.*); A:(\d+\.\d+|NaN); V:(\d+\.\d+|NaN); D:(\d+\.\d+|NaN); N:(\d+\.\d+|NaN);$"  # noqa
)

emotion_map = {
    "A": "anger",
    "H": "happiness",
    "S": "sadness",
    "N": "neutral",
    "O": "other",
    "X": "unknown",
}

ratings_map = {"angry": "anger", "happy": "happiness", "sad": "sadness"}

unused_emotions = ["O", "X"]


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the MSP-IMPROV dataset at location INPUT_DIR and resample
    audio to 16 kHz 16-bit WAV audio.
    """

    dimensions = {}
    labels = {}
    _ratings = []
    with open(input_dir / "Evaluation.txt") as fid:
        name = ""
        for line in fid:
            line = line.strip()
            match = REGEX.match(line)
            match_rater = RATER_RE.match(line)
            if match:
                name = "MSP-IMPROV-" + match.group(1)
                labels[name] = match.group(2)
                dimensions[name] = list(map(float, match.group(3, 4, 5, 6)))
            elif match_rater:
                rater = match_rater.group(1).strip()
                label = match_rater.group(2).strip().lower()
                label = ratings_map.get(label, label)
                # other_labels = match_rater.group(3).strip()
                act = float(match_rater.group(4).strip())
                val = float(match_rater.group(5).strip())
                dom = float(match_rater.group(6).strip())
                nat = float(match_rater.group(7).strip())
                _ratings.append((name, rater, label, act, val, dom, nat))

    paths = list(input_dir.glob("session?/*/*/*.wav"))
    if resample:
        resample_dir = Path("resampled")
        resample_audio(paths, resample_dir)
        paths = list(resample_dir.glob("*.wav"))
    write_filelist(paths, "files_all")
    write_filelist(
        [p for p in paths if labels[p.stem] not in unused_emotions],
        "files_4class",
    )

    write_annotations({n: emotion_map[labels[n]] for n in labels}, "label")
    write_annotations({p.stem: emotion_map[p.stem[14]] for p in paths}, "acted_label")
    write_annotations({p.stem: p.stem[20] for p in paths}, "recording")
    write_annotations({p.stem: p.stem[12:14] for p in paths}, "sentence")
    speaker_dict = {p.stem: p.stem[16:19] for p in paths}
    write_annotations(speaker_dict, "speaker")
    write_annotations({k: v[0] for k, v in speaker_dict.items()}, "gender")
    write_annotations({k: v[-2:] for k, v in speaker_dict.items()}, "session")
    write_annotations({p.stem: "en" for p in paths}, "language")
    write_annotations({p.stem: "us" for p in paths}, "country")

    # Aggregated dimensional annotations per utterance
    df = pd.DataFrame.from_dict(
        dimensions,
        orient="index",
        columns=["activation", "valence", "dominance", "naturalness"],
    )
    df.index.name = "name"
    for dim in ["activation", "valence", "dominance", "naturalness"]:
        write_annotations(df[dim].to_dict(), dim)

    #
    # Ratings analysis
    #
    ratings = pd.DataFrame(
        sorted(_ratings), columns=["name", "rater", "label", "act", "val", "dom", "nat"]
    )
    print("Number of duplicated raters per clip:")
    print(ratings.groupby(["name", "rater"]).size().value_counts().loc[2:])
    # They appear to simply be copied
    ratings.drop_duplicates(["name", "rater"], inplace=True)
    ratings.set_index(["name", "rater"]).to_csv("ratings.csv")

    # Frequency of mode label(s) per utterance
    name_groups = ratings.groupby("name")
    mode_labels = (
        name_groups["label"]
        .value_counts(sort=True, ascending=False)
        .rename("freq")
        .reset_index(level="label")
        .groupby(level="name")
        .first()
        .join(name_groups.size().rename("total"))
    )

    # Include only names with a label which is strictly a plurality
    mode_labels = mode_labels.loc[
        name_groups["label"].agg(lambda x: "".join(x.mode())).isin(set("NASH"))
    ]
    print(f"{len(mode_labels)} clips have plurality vote")

    # Agreement is mean proportion of labels which are plurality label
    agreement = np.mean(mode_labels["freq"] / mode_labels["total"])
    print(f"Mean label agreement: {agreement:.3f}")

    # Acted label
    mode_labels["acted"] = [x[14] for x in mode_labels.index]

    # Agreement with acted label
    agreement = (mode_labels["label"] == mode_labels["acted"]).mean()
    print(f"Plurality vote accuracy to acted: {agreement:.3f}")

    clips = ratings.merge(mode_labels, on="name", suffixes=("_human", "_vote"))
    accuracy = (clips["label_human"] == clips["label_vote"]).mean()
    print(f"Human accuracy to plurality vote: {accuracy:.3f}")
    accuracy = (clips["label_human"] == clips["acted"]).mean()
    print(f"Human accuracy to acted: {accuracy:.3f}")

    # Simple way to get int matrix of labels for raters x clips
    data = (
        ratings.set_index(["rater", "name"])["label"]
        .astype("category")
        .cat.codes.unstack()
        + 1
    )
    data[data.isna()] = 0
    data = data.astype(int).to_numpy()
    print(f"Krippendorf's alpha: {alpha(data):.3f}")


if __name__ == "__main__":
    main()



"""Implementation of the model from [1]_.

References
----------
.. [1] Z. Zhang, B. Wu, and B. Schuller, 'Attention-augmented End-to-end
       Multi-task Learning for Emotion Prediction from Speech', in 2019
       IEEE International Conference on Acoustics, Speech and Signal
       Processing, May 2019, pp. 6705-6709, doi:
       10.1109/ICASSP.2019.8682896.
"""

import torch
from torch import nn

from ertk.pytorch.utils import frame_tensor

from ._base import PyTorchModelConfig, SimpleClassificationModel
from .layers import Attention1D


class Model(SimpleClassificationModel, fname="zhang2019", config=PyTorchModelConfig):
    def __init__(self, config: PyTorchModelConfig) -> None:
        super().__init__(config)

        self.dropout = nn.Dropout(0.1)
        self.conv1 = nn.Sequential(
            nn.Conv1d(1, 40, kernel_size=40, padding=20), nn.ReLU(inplace=True)
        )
        self.maxpool1 = nn.MaxPool1d(2)
        self.conv2 = nn.Sequential(
            nn.Conv1d(40, 40, kernel_size=40, padding=20), nn.ReLU(inplace=True)
        )
        self.maxpool2 = nn.MaxPool1d(10)

        self.gru = nn.GRU(1280, 128, num_layers=2, batch_first=True)
        self.attention = Attention1D(128)
        self.predict = nn.Linear(128, config.n_classes)

    def forward(self, x: torch.Tensor):  # type: ignore
        # x has shape (batch, steps, 640)
        batch_size, steps, _ = x.size()
        x = self.dropout(x).view(batch_size * steps, 1, 640)

        # Padding is even so we need to cut the last value
        extraction = self.conv1(x)[:, :, :-1]
        extraction = self.maxpool1(extraction)
        extraction = self.conv2(extraction)[:, :, :-1]
        # (batch * steps, 40, 320) -> (batch * steps, 320, 40)
        extraction = extraction.transpose(1, 2)
        extraction = self.maxpool2(extraction)
        extraction = extraction.transpose(1, 2).view(batch_size, steps, 1280)

        h0 = torch.zeros(2, batch_size, 128, dtype=x.dtype, device=x.device)
        gru, _ = self.gru(extraction, h0)
        att = self.attention(gru)
        out = self.predict(att)
        return out

    def preprocess_input(self, x: torch.Tensor):
        return frame_tensor(
            x, frame_size=640, frame_shift=160, pad=False, axis=1
        ).squeeze(-1)



"""
TensorFlow backend
==================

This package contains the TensorFlow backend for the ERTK.

.. autosummary::
    :toctree: generated/

    models
    train
    dataset
    classification
    utils
"""

from .models import get_tf_model, get_tf_model_fn
from .utils import compile_wrap, init_gpu_memory_growth, test_fit



"""Process the raw TESS dataset.

This assumes the file structure from the original compressed file:
/.../
    *.wav
"""

from pathlib import Path

import click

from ertk.dataset import resample_audio, write_annotations, write_filelist

emotion_map = {
    "angry": "anger",
    "disgust": "disgust",
    "fear": "fear",
    "happy": "happiness",
    "ps": "surprise",
    "sad": "sadness",
    "neutral": "neutral",
}


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the TESS dataset at location INPUT_DIR and resample audio
    to 16 kHz 16-bit WAV audio.
    """

    paths = list(input_dir.glob("*.wav"))
    if len(paths) == 0:
        raise FileNotFoundError("No audio files found.")
    if resample:
        resample_dir = Path("resampled")
        resample_audio(paths, resample_dir)
        write_filelist(resample_dir.glob("*.wav"), "files_all")

    write_annotations(
        {p.stem: emotion_map[p.stem[p.stem.rfind("_") + 1 :]] for p in paths},
        "label",
    )
    write_annotations({p.stem: p.stem[:3] for p in paths}, "speaker")
    write_annotations({p.stem: "en" for p in paths}, "language")
    write_annotations({p.stem: "ca" for p in paths}, "country")


if __name__ == "__main__":
    main()



"""Dataset utilities for TensorFlow."""

from typing import Callable, List, Optional, Tuple, Union

import numpy as np
import tensorflow as tf
from keras.utils.data_utils import Sequence

from ertk.utils import batch_arrays_by_length, shuffle_multiple

__all__ = [
    "tf_dataset_gen",
    "tf_dataset_mem_ragged",
    "tf_dataset_mem",
    "BatchedSequence",
    "BatchedFrameSequence",
    "DataFunction",
]


DataFunction = Callable[..., tf.data.Dataset]


def tf_dataset_gen(
    x: np.ndarray,
    y: np.ndarray,
    sample_weight: Optional[np.ndarray] = None,
    *,
    batch_size: int = 64,
    shuffle: bool = True,
):
    """Returns a TensorFlow generator Dataset instance with the given
    data.

    Parameters
    ----------
    x: numpy.ndarray
        A 2- or 3-D data matrix of shape (n_instances, n_features) or
        (n_instances, seq_len, n_features).
    y: numpy.ndarray
        A 1-D array of length n_instances containing numeric class
        labels.
    sample_weight: numpy.ndarray, optional
        A 1-D array of length n_instances containing sample weights.
        Added as third item in dataset if present.
    batch_size: int
        The batch size to use.
    shuffle: boolean
        Whether or not to shuffle the dataset. Note that shuffling is
        done *before* batching, unlike in `create_tf_dataset_ragged()`.
    """

    def gen_inst():
        if shuffle:
            perm = np.random.permutation(len(x))
        else:
            perm = np.arange(len(x))

        if sample_weight is None:
            for i in perm:
                yield x[i], y[i]
        else:
            for i in perm:
                yield x[i], y[i], sample_weight[i]

    sig: Tuple[tf.TensorSpec, ...] = (
        tf.TensorSpec(shape=x[0].shape, dtype=tf.float32),
        tf.TensorSpec(shape=(), dtype=tf.int64),
    )
    if sample_weight is not None:
        sig += (tf.TensorSpec(shape=(), dtype=tf.float32),)
    data = tf.data.Dataset.from_generator(gen_inst, output_signature=sig)
    return data.batch(batch_size).prefetch(2)


def tf_dataset_mem(
    x: np.ndarray,
    y: np.ndarray,
    sample_weight: Optional[np.ndarray] = None,
    *,
    batch_size: int = 64,
    shuffle: bool = True,
) -> tf.data.Dataset:
    """Returns a TensorFlow in-memory Dataset instance with the given
    data.

    Parameters
    ----------
    x: numpy.ndarray
        A 2- or 3-D data matrix of shape (n_instances, n_features) or
        (n_instances, seq_len, n_features).
    y: numpy.ndarray
        A 1-D array of length n_instances containing numeric class
        labels.
    sample_weight: numpy.ndarray, optional
        A 1-D array of length n_instances containing sample weights.
        Added as third item in dataset if present.
    batch_size: int
        The batch size to use.
    shuffle: boolean
        Whether or not to shuffle the dataset. Note that shuffling is
        done *before* batching, unlike in `create_tf_dataset_ragged()`.
    """
    with tf.device("CPU"):
        if sample_weight is None:
            data = tf.data.Dataset.from_tensor_slices((x, y))
        else:
            data = tf.data.Dataset.from_tensor_slices((x, y, sample_weight))

    if shuffle:
        data = data.shuffle(len(x))
    return data.batch(batch_size).prefetch(2)


def tf_dataset_mem_ragged(
    x: np.ndarray,
    y: np.ndarray,
    sample_weight: Optional[np.ndarray] = None,
    *,
    batch_size: int = 64,
    shuffle: bool = True,
) -> tf.data.Dataset:
    """Returns a TensorFlow in-memory Dataset instance from
    variable-length features.

    Parameters
    ----------
    x: numpy.ndarray
        A 3-D data matrix of shape (n_instances, length[i], n_features)
        with variable length axis 1.
    y: numpy.ndarray
        A 1-D array of length n_instances containing numeric class
        labels.
    sample_weight: numpy.ndarray, optional
        A 1-D array of length n_instances containing sample weights.
        Added as third item in dataset if present.
    batch_size: int
        The batch size to use.
    shuffle: boolean
        Whether or not to shuffle the dataset. Note that shuffling is
        done *after* batching, because sequences are sorted by length,
        then batched in similar lengths.
    """

    def ragged_to_dense(x: tf.RaggedTensor, y):
        return x.to_tensor(), y

    def ragged_to_dense_weighted(x: tf.RaggedTensor, y, sample_weight):
        return x.to_tensor(), y, sample_weight

    # Sort according to length
    perm = np.argsort([len(a) for a in x])
    x = x[perm]
    y = y[perm]
    if sample_weight is not None:
        sample_weight = sample_weight[perm]

    ragged = tf.RaggedTensor.from_row_lengths(
        np.concatenate(list(x)), [len(a) for a in x]
    )
    with tf.device("CPU"):
        if sample_weight is None:
            data = tf.data.Dataset.from_tensor_slices((ragged, y))
        else:
            data = tf.data.Dataset.from_tensor_slices((ragged, y, sample_weight))

    # Group similar lengths in batches, then shuffle batches
    data = data.batch(batch_size)
    if shuffle:
        data = data.shuffle(len(x) // batch_size + 1)

    if sample_weight is None:
        data = data.map(ragged_to_dense)
    else:
        data = data.map(ragged_to_dense_weighted)
    return data.prefetch(2)


class BatchedFrameSequence(Sequence):
    """Creates a sequence of batches of frames to process.

    Parameters
    ----------
    x: ndarray or list of ndarray
        Sequences of vectors.
    y: ndarray
        Labels corresponding to sequences in x.
    prebatched: bool, default = False
        Whether or not x has already been grouped into batches.
    batch_size: int, default = 32
        Batch size to use. Each generated batch will be at most this size.
    shuffle: bool, default = True
        Whether to shuffle the order of the batches.
    """

    def __init__(
        self,
        x: Union[np.ndarray, List[np.ndarray]],
        y: np.ndarray,
        prebatched: bool = False,
        batch_size: int = 32,
        shuffle: bool = True,
    ):
        self.x = x
        self.y = y
        if not prebatched:
            self.x, self.y = batch_arrays_by_length(
                self.x, self.y, batch_size=batch_size, shuffle=shuffle
            )
        if shuffle:
            self.x, self.y = shuffle_multiple(self.x, self.y, numpy_indexing=True)

    def __len__(self):
        return len(self.x)

    def __getitem__(self, idx: int):
        return self.x[idx], self.y[idx]


class BatchedSequence(Sequence):
    """Creates a sequence of batches to process.

    Parameters
    ----------
    x: ndarray or list of ndarray
        Instance feature vectors. Each vector is assumed to be for a different
        instance.
    y: ndarray
        Labels corresponding to sequences in x.
    prebatched: bool, default = False
        Whether or not x has already been grouped into batches.
    batch_size: int, default = 32
        Batch size to use. Each generated batch will be at most this size.
    shuffle: bool, default = True
        Whether to shuffle the instances.
    """

    def __init__(
        self, x: np.ndarray, y: np.ndarray, batch_size: int = 32, shuffle: bool = True
    ):
        self.x = x
        self.y = y
        self.batch_size = batch_size
        if shuffle:
            self.x, self.y = shuffle_multiple(self.x, self.y, numpy_indexing=True)

    def __len__(self):
        return int(np.ceil(len(self.x) / self.batch_size))

    def __getitem__(self, idx: int):
        sl = slice(idx * self.batch_size, (idx + 1) * self.batch_size)
        return self.x[sl], self.y[sl]


_DATA_FNS = {
    "mem": tf_dataset_mem,
    "gen": tf_dataset_gen,
    "mem_ragged": tf_dataset_mem_ragged,
}


def get_data_fn(name: str) -> Callable[..., tf.data.Dataset]:
    """Returns a data function by name.

    Parameters
    ----------
    name: str
        The name of the data function to return. Can be one of:
        * "``mem``": In-memory data function.
        * "``gen``": Generator data function.
        * "``mem_ragged``": In-memory data function for ragged tensors.

    Returns
    -------
    Callable[..., tf.data.Dataset]
        The data function.
    """
    if ":" in name:
        import importlib

        modname, fname = name.split(":")
        module = importlib.import_module(modname)
        func = getattr(module, fname)
    else:
        func = _DATA_FNS[name]
    return func



"""PyTorch layers for use in models."""

import math
from typing import Any, Callable, List, Sequence

import torch
import torch.nn.functional as F
from torch import nn

from ertk.pytorch.utils import get_activation

__all__ = ["Attention1D", "make_fc"]


class Attention1D(nn.Module):
    """1D attention layer.

    Parameters
    ----------
    in_channels: int
        The number of input channels.
    """

    def __init__(self, in_channels: int) -> None:
        super().__init__()

        self.weight = nn.Parameter(torch.randn(in_channels, 1) / math.sqrt(in_channels))

    def forward(self, x: torch.Tensor):
        # x : (..., length, in_channels)
        alpha = F.softmax(x.matmul(self.weight), -2)
        # alpha : (..., length, 1)
        r = torch.sum(alpha * x, -2)
        return r


def make_fc(
    dims: List[int],
    activation: str = "relu",
    activation_args: Sequence[Any] = [],
    init_fn: Callable = nn.init.xavier_uniform_,
    norm: str = "none",
    dropout: float = 0,
):
    """Makes a fully connected module with the given numbers of hidden
    units, and given activation after every layer except the last.

    Parameters
    ----------
    dims: list of int
        Layer dimensions.
    activation: torch.nn.Module
        The activation to use as a module class.
    activation_param: optional
        A parameter to pass to the activation.
    init_fn: callable
        The weight initialisation scheme. Default is 'xavier_uniform'.
    norm: str
        Whether to apply normalisation after each layer. If "batch" then
        apply batch normalisation. If "layer" apply layer normalisation.

    Returns
    -------
    torch.nn.Module
        A `Module` which contains the corresponding layers, activations,
        and optional dropout and normalisation.
    """
    param = activation_args[0] if len(activation_args) > 0 else None
    init_gain = nn.init.calculate_gain(activation, param=param)

    module = nn.Sequential()
    for i in range(len(dims) - 2):
        linear = nn.Linear(dims[i], dims[i + 1])
        init_fn(linear.weight, gain=init_gain)
        module.append(linear)
        module.append(get_activation(activation, *activation_args))
        if norm == "batch":
            module.append(nn.BatchNorm1d(dims[i + 1]))
        elif norm == "layer":
            module.append(nn.LayerNorm(dims[i + 1]))
        if dropout > 0:
            module.append(nn.Dropout(dropout))
    linear = nn.Linear(dims[-2], dims[-1])
    init_fn(linear.weight, gain=init_gain)
    module.append(linear)
    return module



"""Process the raw Portuguese dataset (Castro & Lima).

This assumes the file structure from the original compressed file:
/.../
    Castro_2010_AppxB_Sents.txt
    Castro_2010_AppxB_Pseudosents.txt
    *.wav
"""

import re
from pathlib import Path

import click
import pandas as pd

from ertk.dataset import resample_audio, write_annotations, write_filelist

REGEX = re.compile(r"^\d+[sp][AB]_([a-z]+)\d+$")

emotion_map = {
    "angry": "anger",
    "disgust": "disgust",
    "fear": "fear",
    "happy": "happiness",
    "sad": "sadness",
    "neutral": "neutral",
    "surprise": "surprise",
}

sentence_map = {
    # Sentences
    "estaMesa": "Esta mesa é de madeira",
    "oRadio": "O rádio está ligado",
    "aqueleLivro": "Aquele livro é de história",
    "aTerra": "A Terra é um planeta",
    "oCao": "O cão trouxe a bola",
    "eleChega": "Ele chega amanhã",
    "estaRoupa": "Esta roupa é colorida",
    "osJardins": "Os jardins têm flores",
    "asPessoas": "As pessoas vão a concertos",
    "haArvores": "Há árvores na floresta",
    "osTigres": "Os tigres são selvagens",
    "oQuadro": "O quadro está na parede",
    "alguemFechou": "Alguém fechou as janelas",
    "osJovens": "Os jovens ouvem música",
    "oFutebol": "O futebol é um desporto",
    "elaViajou": "Ela viajou de comboio",
    # Pseudo-sentences
    "estaDepa": "Esta dêpa é de faneira",
    "oDarrio": "O dárrio está guilado",
    "aqueleJicro": "Aquele jicro é de hisbólia",
    "aPirra": "A Pirra é um flaneto",
    "oLao": "O lão droube a nóma",
    "eleChena": "Ele chena aguinhã",
    "estaSouda": "Esta souda é lacoripa",
    "osBartins": "Os bartins têm pléres",
    "asSemoas": "As semoas vão a cambêrtos",
    "haArjuques": "Há árjuques na plurisca",
    "osLagres": "Os lagres são siltávens",
    "oJuadre": "O juadre está na pafêne",
    "alguemBelhou": "Alguém belhou as jalétas",
    "osDofens": "Os dófens mavem tézica",
    "oDutebel": "O dutebel é um nesforpo",
    "elaJiavou": "Ela jiavou de lantóio",
}


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the Portuguese dataset at location INPUT_DIR and resample
    audio to 16 kHz 16-bit WAV audio.
    """

    paths = list(input_dir.glob("*.wav"))
    if resample:
        resample_dir = Path("resampled")
        resample_audio(paths, resample_dir)
        write_filelist(resample_dir.glob("*.wav"), "files_all")

    sent_info = pd.concat(
        [
            pd.read_csv(
                input_dir / x,
                sep="\t",
                skiprows=5,
                header=0,
                index_col=0,
            )
            for x in [
                "Castro_2010_AppxB_Sents.txt",
                "Castro_2010_AppxB_Pseudosents.txt",
            ]
        ],
        ignore_index=True,
    ).set_index("Stimulus ")

    write_annotations(sent_info["Content"].map(sentence_map), "transcript")
    write_annotations(sent_info["Intensity (1-7)"], "intensity")
    write_annotations(
        {p.stem: emotion_map[REGEX.match(p.stem).group(1)] for p in paths},
        "label",
    )
    write_annotations({p.stem: p.stem[p.stem.find("_") - 1] for p in paths}, "speaker")
    write_annotations(
        {p.stem: p.stem[p.stem.find("_") - 2] for p in paths}, "sent_type"
    )
    write_annotations({p.stem: "pt" for p in paths}, "language")


if __name__ == "__main__":
    main()



import logging
import pprint
from dataclasses import asdict
from functools import partial
from pathlib import Path
from typing import Tuple

import click
from omegaconf import OmegaConf
from tqdm import tqdm

from ertk.dataset import read_features_iterable, write_features
from ertk.utils import TqdmMultiprocessing

logger = logging.getLogger(__name__)


def list_processors(ctx: click.Context, param, val):
    from ertk.preprocessing import InstanceProcessor

    if not val or ctx.resilient_parsing:
        return

    print(f"\nAvailable processors:\n{InstanceProcessor.valid_processors()}\n")
    print(
        "Below are the default configurations for registered feature extractors. Items "
        "with '???' are required to be specified manually.\n"
    )
    for key in sorted(InstanceProcessor.valid_processors()):
        print(f"[{key}]: ", end="")
        proc_cls = InstanceProcessor.get_processor_class(key)
        print(", ".join([b.__name__ for b in proc_cls.__bases__]))
        config_cls = proc_cls.get_config_type()
        for k, v in asdict(config_cls()).items():
            print(f"  {k} = {v}")
        print()

    ctx.exit()


@click.command()
@click.argument("input", type=click.Path(exists=True, path_type=Path))
@click.argument("output", type=Path)
@click.option("--processor", "--features", required=True, help="Processor to apply.")
@click.option(
    "--list_processors",
    is_flag=True,
    callback=list_processors,
    expose_value=False,
    is_eager=True,
    help="Show options for registered processors.",
)
@click.option(
    "--config",
    type=click.Path(exists=True, path_type=Path),
    help="Extractor config file.",
)
@click.option("--corpus", default="", help="Corpus name to set.")
@click.option(
    "--batch_size",
    type=int,
    default=1,
    help="Batch size for processing. If batch_size is greater than 1, clips will be "
    "batched together.",
)
@click.option(
    "--sample_rate",
    type=int,
    default=16000,
    help="Resample to this rate for audio input.",
)
@click.option("--n_jobs", type=int, default=1, help="Number of parallel jobs to run.")
@click.option("--verbose", type=int, default=0)
@click.argument("restargs", nargs=-1)
def main(
    input: Path,
    output: Path,
    processor: str,
    config: Path,
    corpus: str,
    batch_size: int,
    sample_rate: int,
    n_jobs: int,
    verbose: int,
    restargs: Tuple[str],
):
    """Process features or audio files in INPUT, write to OUTPUT."""

    from ertk.preprocessing import InstanceProcessor

    if verbose > 0:
        logging.basicConfig(level=logging.DEBUG if verbose > 1 else logging.INFO)

    extractor_cls = InstanceProcessor.get_processor_class(processor)
    config_cls = extractor_cls.get_config_type()
    if config:
        conf = OmegaConf.load(config)
    else:
        conf = extractor_cls.get_default_config()
    conf = OmegaConf.merge(conf, OmegaConf.from_dotlist(list(restargs)))
    logger.info("Using configuration:")
    logger.info(pprint.pformat(dict(conf)))
    extractor = extractor_cls(config_cls.from_config(conf))

    input_data = read_features_iterable(input, sample_rate=sample_rate)
    corpus = corpus or input_data.corpus
    names = input_data.names
    if batch_size != 1:
        feats = tqdm(
            extractor.process_all(iter(input_data), batch_size, sr=sample_rate),
            total=len(input_data),
            desc="Processing files",
            disable=None,
        )
    else:
        # TODO: Set back to joblib when
        # https://github.com/joblib/joblib/pull/588 is merged
        pool = TqdmMultiprocessing(len(input_data), "Processing files", n_jobs=n_jobs)
        f = partial(extractor.process_instance, sr=sample_rate)
        feats = pool.imap(f, input_data)
    write_features(
        output, feats, names=names, corpus=corpus, feature_names=extractor.feature_names
    )
    extractor.finish()
    print(f"Wrote features to {output}")



"""
Training and evaluation classes and functions
=============================================

Config classes
--------------
.. autosummary::
    :toctree: generated/

    TrainConfig
    ExperimentResult
    ExperimentConfig
    ModelConfig
    CrossValidationConfig
    EvalConfig


Splitting classes and functions
-------------------------------
.. autosummary::
    :toctree: generated/

    TrainValidation
    ShuffleGroupKFold
    ValidationSplit
    get_cv_splitter


Miscellaneous functions
-----------------------
.. autosummary::
    :toctree: generated/

    get_pipeline_params
    get_scores
    scores_to_df
"""

from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Type, TypeVar, Union

import numpy as np
import omegaconf
import pandas as pd
from sklearn.base import MetaEstimatorMixin
from sklearn.metrics import get_scorer
from sklearn.model_selection import (
    BaseCrossValidator,
    GroupKFold,
    GroupShuffleSplit,
    LeaveOneGroupOut,
    LeaveOneOut,
    StratifiedKFold,
    StratifiedShuffleSplit,
)
from sklearn.model_selection._split import _BaseKFold
from sklearn.model_selection._validation import _score
from sklearn.pipeline import Pipeline
from sklearn.utils import check_array

from ertk.config import ERTKConfig
from ertk.dataset.dataset import DataLoadConfig, DataSelector
from ertk.utils import PathOrStr, ScoreFunction, filter_kwargs

__all__ = [
    "TrainConfig",
    "ExperimentResult",
    "ExperimentConfig",
    "ModelConfig",
    "CrossValidationConfig",
    "EvalConfig",
    "TrainValidation",
    "ShuffleGroupKFold",
    "ValidationSplit",
    "get_cv_splitter",
    "get_pipeline_params",
    "get_scores",
    "scores_to_df",
]


class TrainValidation(BaseCrossValidator):
    """Validation method that uses the training set as validation set."""

    def split(self, X, y, groups):
        yield np.arange(len(X)), np.arange(len(X))

    def get_n_splits(self, X, y, groups):
        return 1


class ShuffleGroupKFold(_BaseKFold):
    """Like GroupKFold but with random combinations of groups instead of
    deterministic combinations based on group size. This is most useful
    if you have groups of near equal size, and you want group k-fold CV,
    where k divides n_groups.

    Note: If shuffle=False, this does not behave identical to
    GroupKFold, but rather splits groups in sorted order (as returned by
    `numpy.unique()`).
    """

    def __init__(self, n_splits=5, *, shuffle=False, random_state=None):
        super().__init__(n_splits=n_splits, shuffle=shuffle, random_state=random_state)

    def _iter_test_indices(self, X, y, groups):
        if groups is None:
            raise ValueError("The 'groups' parameter should not be None.")
        groups = check_array(groups, ensure_2d=False, dtype=None)

        unique_groups, groups = np.unique(groups, return_inverse=True)
        n_groups = len(unique_groups)

        if self.n_splits > n_groups:
            raise ValueError(
                "Cannot have number of splits n_splits=%d greater"
                " than the number of groups: %d." % (self.n_splits, n_groups)
            )

        # Pairs of start,end indices of groups each of n folds
        fold_idx = np.linspace(0, n_groups, self.n_splits + 1, dtype=int)

        group_order = np.arange(n_groups)
        if self.shuffle:
            # Shuffle order groups appear in folds
            group_order = np.random.default_rng(self.random_state).permutation(
                group_order
            )

        # Mapping from group index to fold index
        group_to_fold = np.zeros(n_groups)

        for fold, (g1, g2) in enumerate(zip(fold_idx[:-1], fold_idx[1:])):
            group_to_fold[group_order[g1:g2]] = fold

        indices = group_to_fold[groups]

        for f in range(self.n_splits):
            yield np.where(indices == f)[0]


class ValidationSplit(BaseCrossValidator):
    """Validation method that uses a pre-defined validation set.

    Parameters
    ----------
    train_idx: list of int or np.ndarray
        Indices of the training set.
    valid_idx: list of int or np.ndarray
        Indices of the validation set.
    """

    def __init__(
        self,
        train_idx: Union[List[int], np.ndarray],
        valid_idx: Union[List[int], np.ndarray],
    ):
        self.train_idx = train_idx
        self.valid_idx = valid_idx

    def split(self, X, y, groups):
        if self.train_idx is not None:
            train_idx = np.array(self.train_idx)
        else:
            train_idx = np.arange(len(X))
            train_idx = train_idx[~np.isin(train_idx, self.valid_idx)]
        yield train_idx, self.valid_idx

    def get_n_splits(self, X, y, groups):
        return 1


def get_cv_splitter(
    group: bool,
    k: int,
    test_size: float = 0.2,
    shuffle: bool = False,
    random_state: int = None,
) -> BaseCrossValidator:
    """Gets an appropriate cross-validation splitter for the given
    number of folds and groups, or a single random split.

    Parameters
    ----------
    group: bool
        Whether to split over pre-defined groups of instances.
    k: int
        If k > 1 then do k-fold CV. If k == 1 then do one random
        split. If k = -1 then do leave-one-out. If k == 0 then use the
        whole train set as validation split.
    test_size: float
        The size of the test set when k == 1 (one random split).
    shuffle: bool
        Whether to shuffle when using k-fold for k > 1.
    random_state: int, optional
        The random state to set for splitters with shuffling behaviour.

    Returns
    -------
    splitter: BaseCrossValidator
        Cross-validation splitter that has `split()` and
        `get_n_splits()` methods.
    """
    # TODO: Leave-|k|-out for k < 0?
    if k == 0:
        return TrainValidation()
    if group:
        if k > 1:
            if shuffle:
                return ShuffleGroupKFold(k, shuffle=shuffle, random_state=random_state)
            return GroupKFold(k)
        elif k == 1:
            return GroupShuffleSplit(1, test_size=test_size, random_state=random_state)
        return LeaveOneGroupOut()
    if k > 1:
        return StratifiedKFold(k, shuffle=shuffle, random_state=random_state)
    elif k == 1:
        return StratifiedShuffleSplit(1, test_size=test_size, random_state=random_state)
    return LeaveOneOut()


def get_pipeline_params(params: Dict[str, Any], pipeline: Pipeline):
    """Modifies parameter names to pass to a Pipeline instance's `fit()`
    method.

    Parameters
    ----------
    params: dict
        Parameters to pass to Pipeline.fit(). All parameters are passed
        to all estimators in the pipeline so long as they are valid.
    pipeline: Pipeline
        The pipeline instance.

    Returns
    -------
    new_params: dict
        Parameters filtered and prepended with pipeline step names and
        double underscore (e.g. groups -> clf__groups).
    """
    new_params = {}
    for name, est in pipeline.named_steps.items():
        if est is None or est == "passthrough":
            continue
        filt_params = params
        if not isinstance(est, MetaEstimatorMixin):  # May pass through kwargs
            filt_params = filter_kwargs(filt_params, est.fit)
        new_params.update({name + "__" + k: v for k, v in filt_params.items()})
    return new_params


def get_scores(
    scoring: Union[str, List[str], Dict[str, ScoreFunction], Callable[..., float]],
    y_pred: np.ndarray,
    y_true: np.ndarray,
) -> Dict[str, Any]:
    """Get dictionary of scores for predictions.

    Parameters
    ----------
    scoring: str, list, dict or callable
        Score(s) to calculate. This takes the same for as for
        scikit-learn's cross_val_* methods.
    y_pred: array-like
        Predictions.
    y_true: array-like
        Ground truth.

    Returns
    -------
    scores: dict
        A dictionary mapping score names to corresponding score(s).
    """

    class DummyEstimator:
        """Class that implements a dummy estimator for scoring, to avoid
        repeated invocations of `predict()` etc.
        """

        def __init__(self, y_pred):
            self.y_pred = y_pred

        def predict(self, x, **kwargs):
            return self.y_pred

        def predict_proba(self, x, **kwargs):
            return self.y_pred

        def decision_function(self, x, **kwargs):
            return self.y_pred

    y_pred = np.array(y_pred)
    y_true = np.array(y_true)
    dummy = DummyEstimator(y_pred)

    if isinstance(scoring, str):
        scoring = {scoring: get_scorer(scoring)}
    elif callable(scoring):
        scoring = {"score": scoring}
    elif isinstance(scoring, list):
        scoring = {x: get_scorer(x) for x in scoring}
    return _score(dummy, None, y_true, scoring)


def scores_to_df(
    scores: Dict[str, np.ndarray], index: Union[pd.Index, Iterable, None] = None
):
    """Convert scikit-learn scores dictionary to pandas dataframe.

    Parameters
    ----------
    scores: dict
        Scores dictionary. Each key is a string and all values are
        either scalar or arrays.
    index: pd.Index or iterable, optional
        Optional index to use. Must have the same length as each value
        in `scores`. If not given a `pd.RangeIndex` is used.

    Returns
    -------
    pd.DataFrame
        Scores dataframe, with a column per key in `scores` and row per
        item in each value array.
    """
    _val = next(iter(scores.values()))
    if index is None:
        if np.ndim(_val) == 0:
            index = pd.RangeIndex(1, name="fold")
        else:
            index = pd.RangeIndex(len(_val), name="fold")
    elif not isinstance(index, pd.Index):
        index = pd.Index(index, name="fold")
    return pd.DataFrame(
        {k[5:] if k.startswith("test_") else k: v for k, v in scores.items()},
        index=index,
    )


@dataclass
class ExperimentResult:
    """Class to hold results of an experiment."""

    scores_dict: Dict[str, np.ndarray] = field(default_factory=dict)
    """Dictionary of scores. Each key is a string and all values are
    either scalar or arrays.
    """
    scores_df: pd.DataFrame = field(default_factory=pd.DataFrame)
    """Scores dataframe, with a column per key in `scores_dict` and row
    per evaluation.
    """
    predictions: Optional[np.ndarray] = None
    """Array of predictions. If not given, this is None."""


@dataclass
class ModelConfig(ERTKConfig):
    """Class to hold model configuration."""

    type: str = omegaconf.MISSING
    """Model type. This is used to determine which model class to
    instantiate.
    """
    config: Any = None
    """Model configuration. This is passed to the model class
    constructor.
    """
    args: Dict[str, Any] = field(default_factory=dict)
    """Model arguments. These are passed to the model class constructor
    as keyword arguments.
    """
    args_path: Optional[str] = None
    """Path to a YAML file containing model arguments. These are passed
    to the model class constructor as keyword arguments.
    """
    param_grid: Dict[str, Any] = field(default_factory=dict)
    """Parameter grid for hyperparameter search. This is passed to
    scikit-learn's `GridSearchCV` class.
    """
    param_grid_path: Optional[str] = None
    """Path to a YAML file containing a parameter grid for
    hyperparameter search. This is passed to scikit-learn's
    `GridSearchCV` class.
    """


@dataclass
class CrossValidationConfig(ERTKConfig):
    """Class to hold cross-validation configuration."""

    part: Optional[str] = None
    """Partition to use for cross-validation. If not given, then
    cross-validation is performed over instances instead of groups.
    """
    kfold: int = omegaconf.MISSING
    """Number of folds for cross-validation."""
    test_size: float = 0.2
    """Fraction of data to use for testing. This is only used if `kfold`
    is 1.
    """


@dataclass
class EvalConfig(ERTKConfig):
    """Class to hold evaluation configuration."""

    cv: Optional[CrossValidationConfig] = None
    """Cross-validation configuration. If not given, then no
    cross-validation is performed.
    """
    train: Optional[DataSelector] = None
    """Data selector for training data."""
    valid: Optional[DataSelector] = None
    """Data selector for validation data."""
    test: Optional[DataSelector] = None
    """Data selector for testing data."""
    inner_kfold: Optional[int] = None
    """Number of folds for inner cross-validation, if using
    hyperparameter search.
    """
    inner_part: Optional[str] = None
    """Partition to use for inner cross-validation, if using
    hyperparameter search.
    """


class TransformClass(Enum):
    std = "std"
    minmax = "minmax"


@dataclass
class TrainConfig(ERTKConfig):
    """Class to hold training configuration."""

    balanced: bool = True
    """Whether to use class-balanced sample weights."""
    reps: int = 1
    """Number of repetitions."""
    normalise: str = "online"
    """Normalisation method. Can be one of "online", or "none"."""
    transform: TransformClass = TransformClass.std
    """Transform method for normalisation. Can be one of "std",
    "minmax", or "none".
    """
    seq_transform: str = "global"
    """Transform method for sequence normalisation."""
    n_jobs: int = 1
    """Number of jobs to run in parallel."""
    verbose: int = 0
    """Verbosity level."""
    label: str = "label"
    """Name of label column."""
    sklearn: Any = None
    """Scikit-learn configuration."""
    pytorch: Any = None
    """PyTorch configuration."""
    tensorflow: Any = None
    """TensorFlow configuration."""


T = TypeVar("T", bound="ExperimentConfig")


@dataclass
class ExperimentConfig(ERTKConfig):
    """Class to hold experiment configuration."""

    name: str = "default"
    """Experiment name."""
    data: DataLoadConfig = omegaconf.MISSING
    """Data loading configuration."""
    model: ModelConfig = omegaconf.MISSING
    """Model configuration."""
    eval: Optional[EvalConfig] = None
    """Evaluation configuration."""
    evals: Dict[str, EvalConfig] = field(default_factory=dict)
    """Dictionary of evaluation configurations."""
    training: TrainConfig = field(default_factory=TrainConfig)
    """Training configuration."""
    results: str = ""
    """Path to output results."""

    @classmethod
    def from_file(
        cls: Type[T], path: PathOrStr, override: Optional[List[str]] = None
    ) -> T:
        path = Path(path)
        conf = super().from_file(path, override)
        if isinstance(conf.data, str):
            conf.data = DataLoadConfig.from_file(path.parent / conf.data)
        return conf



"""Process the raw EmoryNLP dataset.

This assumes the following file structure from the original compressed files:
/.../
    emorynlp_dev_splits/
        *.mp4
    emorynlp_test_splits/
        *.mp4
    emorynlp_train_splits/
        *.mp4
    *.csv
"""

from pathlib import Path

import click
import pandas as pd

from ertk.dataset import resample_audio, write_annotations, write_filelist

NAME_FMT = "sea{0.Season}_ep{0.Episode}_sc{0.Scene_ID}_utt{0.Utterance_ID}"

emotion_map = {
    "Joyful": "joy",
    "Mad": "anger",
    "Neutral": "neutral",
    "Peaceful": "peace",
    "Powerful": "power",
    "Sad": "sadness",
    "Scared": "fear",
}


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the EmoryNLP dataset at location INPUT_DIR and resample
    audio to 16 kHz 16-bit WAV audio.
    """
    splits = ["train", "dev", "test"]
    csvs = [f"emorynlp_{x}_final.csv" for x in splits]

    paths = list(input_dir.glob("**/*.mp4"))
    resample_dir = Path("resampled")
    if resample:
        resample_audio(paths, resample_dir)

    dfs = []
    for split, x in zip(splits, csvs):
        df = pd.read_csv(input_dir / x)
        df.index = df.apply(lambda x: NAME_FMT.format(x), axis=1)
        df["split"] = split
        if split == "train":
            # Corrupted
            df = df.drop(
                ["sea4_ep4_sc3_utt10", "sea4_ep4_sc3_utt8", "sea4_ep4_sc3_utt9"]
            )
        write_filelist({resample_dir / f"{x}.wav" for x in df.index}, f"files_{split}")
        dfs.append(df)
    df = pd.concat(dfs)
    df["Emotion"] = df["Emotion"].map(emotion_map)
    write_filelist({resample_dir / f"{x}.wav" for x in df.index}, "files_all")

    write_annotations(df["Emotion"], "label")
    write_annotations(df["Speaker"], "speaker")
    write_annotations(df["split"], "split")
    write_annotations(df["Utterance"], "transcript")
    write_annotations({x: "us" for x in df.index}, "country")
    write_annotations({x: "en" for x in df.index}, "language")


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import List, Optional

import arff
import netCDF4
import numpy as np
import pandas as pd
import pytest

from ertk.dataset.features import FeaturesData, read_features, write_features

from .constants import (
    all_clips_names,
    corpus_name,
    feature_names,
    features_2d,
    features_3d,
    features_dir,
    features_vlen,
    resample_dir,
    slices_vlen,
)


@pytest.mark.parametrize("format", [".nc", ".csv", ".arff"])
def test_read_features_2d(format: str):
    features_path = features_dir / ("features_2d" + format)
    data = read_features(features_path)
    if format != ".csv":
        assert data.corpus == "test_corpus"
    assert data.feature_names == [f"test_feat{i}" for i in range(15)]
    assert data.names == sorted(x.stem for x in resample_dir.glob("*"))
    assert data.flat.shape == (12, 15)
    assert data.flat.flags.forc
    assert data.features.shape == (12, 15)
    assert data.features is data.flat
    assert list(data.slices) == [1] * 12
    assert np.allclose(data.features, features_2d)


@pytest.mark.parametrize("format", [".nc", ".csv", ".arff"])
def test_read_features_3d(format: str):
    features_path = features_dir / ("features_3d" + format)
    data = read_features(features_path)
    if format != ".csv":
        assert data.corpus == "test_corpus"
    assert data.feature_names == [f"test_feat{i}" for i in range(15)]
    assert data.names == sorted(x.stem for x in resample_dir.glob("*"))
    assert data.flat.shape == (120, 15)
    assert data.flat.flags.forc
    assert data.features.shape == (12, 10, 15)
    assert list(data.slices) == [10] * 12
    assert np.allclose(data.features, features_3d)


@pytest.mark.parametrize("format", [".nc", ".csv", ".arff"])
def test_read_features_vlen(format: str):
    features_path = features_dir / ("features_vlen" + format)
    data = read_features(features_path)
    if format != ".csv":
        assert data.corpus == "test_corpus"
    assert data.feature_names == [f"test_feat{i}" for i in range(15)]
    assert data.names == sorted(x.stem for x in resample_dir.glob("*"))
    assert data.flat.shape == (60, 15)
    assert data.flat.flags.forc
    assert data.features.shape == (12,)
    assert list(data.slices) == [3, 2, 9, 9, 4, 4, 7, 4, 6, 2, 7, 3]
    assert np.allclose(data.flat, features_vlen)


@pytest.mark.parametrize(
    ["name", "features", "slices"],
    [
        ("features_2d", features_2d, None),
        ("features_3d", features_3d, None),
        ("features_vlen", features_vlen, slices_vlen),
    ],
)
@pytest.mark.parametrize("format", [".nc", ".csv", ".arff"])
def test_write_features(
    format: str,
    name: str,
    features: np.ndarray,
    slices: Optional[List[int]],
    tmp_path: Path,
):
    filename = name + format
    ref_path = features_dir / filename
    out_path = tmp_path / filename
    write_features(
        out_path,
        features,
        all_clips_names,
        corpus=corpus_name,
        feature_names=feature_names,
        slices=slices,
    )
    if format == ".csv":
        df1 = pd.read_csv(out_path, header=0, index_col=0)
        df2 = pd.read_csv(ref_path, header=0, index_col=0)
        assert df1.equals(df2)
    elif format == ".arff":
        with open(out_path) as fid:
            data = arff.load(fid)
        with open(ref_path) as fid:
            ref_data = arff.load(fid)
        assert data == ref_data
    elif format == ".nc":
        with netCDF4.Dataset(out_path) as data, netCDF4.Dataset(ref_path) as ref_data:
            assert np.array_equal(data.variables["data"], ref_data.variables["data"])
            assert np.array_equal(data.variables["name"], ref_data.variables["name"])
            assert np.array_equal(
                data.variables["feature"], ref_data.variables["feature"]
            )
            assert np.array_equal(
                data.variables["slices"], ref_data.variables["slices"]
            )
            assert data.corpus == ref_data.corpus


def test_featuresdata_flat():
    data = FeaturesData(np.array([[0.1, 0.2], [0.2, 0.1]]), ["test1", "test2"])
    assert hasattr(data, "flat")


@pytest.mark.parametrize("names", [["test1", "test2"], ["name2", "name1"]])
def test_featuresdata_names(names: List[str]):
    data = FeaturesData(np.array([[0.1, 0.2], [0.2, 0.1]]), names)
    assert data.names == names
    assert data.names is not names


@pytest.mark.parametrize("feature_names", [["f", "g"], ["f1", "f2"], None])
def test_featuresdata_feature_names(feature_names: Optional[List[str]]):
    data = FeaturesData(
        np.array([[0.1, 0.2], [0.2, 0.1]]),
        names=["test1", "test2"],
        feature_names=feature_names,
    )
    if feature_names is not None:
        assert data.feature_names == feature_names
    else:
        assert data.feature_names == [f"feature{i + 1}" for i in range(2)]
    assert data.feature_names is not feature_names


def test_featuresdata_2d():
    data = FeaturesData(features_2d, all_clips_names)
    data.make_contiguous()
    assert len(data.flat.shape) == 2
    assert np.array_equal(data.features, data.flat)
    assert len(data.features) == len(all_clips_names)
    assert len(data.slices) == len(all_clips_names)
    assert list(data.slices) == [1] * len(all_clips_names)


def test_featuresdata_3d():
    n_features = features_3d.shape[2]
    data = FeaturesData(features_3d, all_clips_names)
    data.make_contiguous()
    assert len(data.flat.shape) == 2
    assert len(data.features.shape) == 3
    assert np.array_equal(data.features.reshape(-1, n_features), data.flat)
    assert len(data.features) == len(all_clips_names)
    assert list(data.slices) == [features_3d.shape[1]] * len(all_clips_names)


def test_featuresdata_vlen():
    data = FeaturesData(features_vlen, all_clips_names, slices=slices_vlen)
    data.make_contiguous()
    assert len(data.flat.shape) == 2
    assert data.features.dtype == object
    assert len(data.features) == len(all_clips_names)
    assert list(data.slices) == slices_vlen
